from __future__ import annotations

import copy
import math
import string
from dataclasses import dataclass
from itertools import combinations

import nltk

from plagdef.model.preprocessing import Document, Preprocessor
from plagdef.model.seeding import SentenceMatcher


def sum_vect(dic1, dic2):
    """
    DESCRIPTION: Adding two vectors in form of dictionaries (sparse vectors or inverted list)
    INPUTS: dic1 <dictionary> - Vector 1
            dic2 <dictionary> - Vector 2
    OUTPUT: res <dictionary> - Sum of the two vectors
    """
    res = copy.deepcopy(dic1)
    for i in dic2.keys():
        if i in res:
            res[i] += dic2[i]
        else:
            res[i] = dic2[i]
    return res


def eucl_norm(d1):
    """
    DESCRIPTION: Compute the Euclidean norm of a sparse vector
    INPUT: d1 <dictionary> - sparse vector representation
    OUTPUT: Norm of the sparse vector d1
    """
    norm = 0.0
    for val in d1.values():
        norm += float(val * val)
    return math.sqrt(norm)


def cosine_measure(d1, d2):
    """
    DESCRIPTION: Compute the cosine measure (cosine of the angle between two vectors) in sparse (dictionary)
    representation
    INPUT: d1 <dictionary> - Sparse vector 1
           d2 <dictionary> - Sparse vector 2
    OUTPUT: Cosine measure
    """
    dot_prod = 0.0
    det = eucl_norm(d1) * eucl_norm(d2)
    if det == 0:
        return 0
    for word in d1.keys():
        if word in d2:
            dot_prod += d1[word] * d2[word]
    return dot_prod / det


def adjacent_sents(a, b, th):
    """
    DESCRIPTION: Define if two sentences are adjacent measured in sentences
    INPUT: a <int> - Sentence a index,
           b <int> - Sentence b index
           th <int> - maximum gap between indexes
    OUTPUT: True if the two sentences are adjacents, False otherwise
    """
    if abs(a - b) - 1 <= th:
        return True
    else:
        return False


def frag_founder(ps, src_gap, susp_gap, src_size, susp_size, side):
    """
    DESCRIPTION: Form clusters by grouping "adjacent" sentences in a given side (source o suspicious)
    INPUT: ps <list of tuples (int, int)> - Seeds
           src_offsets <list of tuples (int, int)> - Contain the char offset and length of each source document sentence
           susp_offsets <list of tuples (int, int)> - Contain the char offset and length of each suspicious document
           sentence
           src_gap <int> - Max gap between sentences to be consider adjacent in the source document
           susp_gap <int> - Max gap between sentences to be consider adjacent in the suspicious document
           src_size <int> - Minimum amount of sentences in a plagiarism case in the side of source document
           susp_size <int> - Minimum amount of sentences in a plagiarism case in the side of suspicious document
           side <0 or 1> 0: Suspicious document side, 1: Source document side
    OUTPUT: res <list of list of tuples (int, int)> - Contains the clusters
    """
    if side == 0:
        max_gap = susp_gap
        min_size = susp_size
    else:
        max_gap = src_gap
        min_size = src_size
    res = []
    ps.sort(key=lambda tup: tup[side])
    sub_set = []
    for pair in ps:
        if len(sub_set) == 0:
            sub_set.append(pair)
        else:
            if adjacent_sents(pair[side], sub_set[-1][side], max_gap):
                sub_set.append(pair)
            else:
                if len(sub_set) >= min_size:
                    res.append(sub_set)
                sub_set = [pair]
    if len(sub_set) >= min_size:
        res.append(sub_set)
    return res


def clustering(ps, src_offsets, susp_offsets, src_gap, susp_gap, src_size, susp_size, side, times):
    """
    DESCRIPTION: Generates the clusters of seeds
    INPUT: ps <list of tuples (int, int)> - Seeds
           src_offsets <list of tuples (int, int)> - Contain the char offset and length of each source document sentence
           susp_offsets <list of tuples (int, int)> - Contain the char offset and length of each suspicious document
           sentence
           src_gap <int> - Max gap between sentences to be consider adjacent in the source document
           susp_gap <int> - Max gap between sentences to be consider adjacent in the suspicious document
           src_size <int> - Minimum amount of sentences in a plagiarism case in the side of source document
           susp_size <int> - Minimum amount of sentences in a plagiarism case in the side of suspicious document
           side <0 or 1> 0: Suspicious document side, 1: Source document side
           times <int> - Counts how many times clustering() have been called
    OUTPUT: res <list of list of tuples (int, int)> - Contains the clusters
    """
    ps_sets = frag_founder(ps, src_gap, susp_gap, src_size, susp_size, side)
    res = []
    if len(ps_sets) <= 1 and times > 0:
        return ps_sets
    else:
        times += 1
        for i in ps_sets:
            partial_res = clustering(i, src_offsets, susp_offsets, src_gap, susp_gap, src_size, susp_size,
                                     (side + 1) % 2, times)
            res.extend(partial_res)
    return res


def validation(plags, psr, src_offsets, susp_offsets, src_bow, susp_bow, src_gap, src_gap_least, susp_gap,
               susp_gap_least, src_size, susp_size, th3):
    """
    DESCRIPTION: Compute the similarity of the resulting plagiarism cases from extension. In case of being below
    certain threshold extension is applied again with max_gap - 1
    INPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Have the plagiarism cases represented by min
    and max sentence index in suspicious and source document respectively
           psr <list of list of tuples (int, int)> - Contains the clusters
           src_offsets <list of tuples (int, int)> - Contain the char offset and length of each source document sentence
           susp_offsets <list of tuples (int, int)> - Contain the char offset and length of each suspicious document
           sentence
           src_bow <list of dictionaries> - Bag of words representing each sentence vector of source document
           susp_bow <list of dictionaries> - Bag of words representing each sentence vector of suspicious document
           src_gap <int> - Max gap between sentences to be consider adjacent in the source document
           src_gap_least <int> - Smallest value the max gap between sentences considerd adjacent can gets in the
           source document
           susp_gap <int> - Max gap between sentences to be consider adjacent in the suspicious document
           susp_gap_least <int> - Smallest value the max gap between sentences considerd adjacent can gets in the
           suspicious document
           src_size <int> - Minimum amount of sentences in a plagiarism case in the side of source document
           susp_size <int> - Minimum amount of sentences in a plagiarism case in the side of suspicious document
           th3 <float> - Threshold for the minimum cosine similarity between source and suspicios fragments in a
           plagiarism case
    OUTPUT: res_plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases that
    passed the validation process
            res_psr <list of list of tuples (int, int)> - Contains the clusters that passed the validation process
            res_sim_frag <list of floats> - Stores the cosine similarity between source and suspicios fragments in
            the plagiarism cases
    """
    res_plags = []
    res_psr = []
    res_sim_frag = []
    i = 0
    range_i = len(plags)
    while i < range_i:
        susp_d = {}
        for j in range(plags[i][0][0], plags[i][0][1] + 1):
            susp_d = sum_vect(susp_d, susp_bow[j])
        src_d = {}
        for j in range(plags[i][1][0], plags[i][1][1] + 1):
            src_d = sum_vect(src_d, src_bow[j])
        sim_frag = cosine_measure(src_d, susp_d)
        if sim_frag <= th3:
            # Did not pass with src_gap
            if src_gap > src_gap_least and susp_gap > susp_gap_least:  # Do until substraction +1
                new_psr = clustering(psr[i], src_offsets, susp_offsets, src_gap - 1, susp_gap - 1, src_size, susp_size,
                                     0, 0)
                new_plags = []
                for ps_set in new_psr:
                    new_plags.append([(min([x[0] for x in ps_set]), max([x[0] for x in ps_set])),
                                      (min([x[1] for x in ps_set]), max([x[1] for x in ps_set]))])
                if len(new_plags) == 0:
                    return []
                temp_res = validation(new_plags, new_psr, src_offsets, susp_offsets, src_bow, susp_bow, src_gap - 1,
                                      src_gap_least, susp_gap - 1, susp_gap_least, src_size, susp_size, th3)
                if len(temp_res) == 0:
                    plags_rec, psr_rec, res_sim_frag_rec = [], [], []
                else:
                    plags_rec, psr_rec, res_sim_frag_rec = temp_res[0], temp_res[1], temp_res[2]
                if len(plags_rec) != 0:
                    res_plags.extend(plags_rec)
                    res_psr.extend(psr_rec)
                    res_sim_frag.extend(res_sim_frag_rec)
            i += 1
        else:
            # Passed with src_gap
            res_plags.append(plags[i])
            res_psr.append(psr[i])
            res_sim_frag.append(sim_frag)
            i += 1
    return res_plags, res_psr, res_sim_frag


def remove_overlap3(plags, psr, src_bow, susp_bow):
    """
    DESCRIPTION: From a set of overlapping plagiarism cases, looking only on the suspicious side, selects the best
    case. See article (1) at the beggining of this file, for the formal description.
    INPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Have the plagiarism cases represented by min
    and max sentence index in suspicious and source document respectively
           psr <list of list of tuples (int, int)> - Contains the clusters
           src_bow <list of dictionaries> - Bag of words representing each sentence vector of source document
           susp_bow <list of dictionaries> - Bag of words representing each sentence vector of suspicious document
    OUTPUT: res_plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases without
    overlapping
            res_psr <list of list of tuples (int, int)> - Contains the clusters without overlapping
    """
    if len(plags) != 0:
        plags, psr = map(list, zip(*sorted(zip(plags, psr), key=lambda tup: tup[0][0][0])))
    res_plags = []
    res_psr = []
    flag = 0
    i = 0
    while i < len(plags):
        cont_ol = 0
        if flag == 0:
            for k in range(i + 1, len(plags)):
                if plags[k][0][0] - plags[i][0][1] <= 0:
                    cont_ol += 1
        else:
            for k in range(i + 1, len(plags)):
                if plags[k][0][0] - res_plags[-1][0][1] <= 0:
                    cont_ol += 1
        if cont_ol == 0:
            if flag == 0:
                res_plags.append(plags[i])
                res_psr.append(psr[i])
            else:
                flag = 0
            i += 1
        else:
            ind_max = i
            higher_sim = 0.0
            for j in range(1, cont_ol + 1):
                if flag == 0:
                    sents_i = range(plags[i][0][0], plags[i][0][1] + 1)
                    range_i = range(plags[i][1][0], plags[i][1][1] + 1)
                else:
                    sents_i = range(res_plags[-1][0][0], res_plags[-1][0][1] + 1)
                    range_i = range(res_plags[-1][1][0], res_plags[-1][1][1] + 1)
                sents_j = range(plags[i + j][0][0], plags[i + j][0][1] + 1)
                sim_i_ol = 0.0
                sim_j_ol = 0.0
                sim_i_nol = 0.0
                sim_j_nol = 0.0
                cont_ol_sents = 0
                cont_i_nol_sents = 0
                cont_j_nol_sents = 0
                for sent in sents_i:
                    sim_max = 0.0
                    for k in range_i:
                        sim = cosine_measure(susp_bow[sent], src_bow[k])
                        if sim > sim_max:
                            sim_max = sim
                    if sent in sents_j:
                        sim_i_ol += sim_max
                        cont_ol_sents += 1
                    else:
                        sim_i_nol += sim_max
                        cont_i_nol_sents += 1
                range_j = range(plags[i + j][1][0], plags[i + j][1][1] + 1)
                for sent in sents_j:
                    sim_max = 0.0
                    for k in range_j:
                        sim = cosine_measure(susp_bow[sent], src_bow[k])
                        if sim > sim_max:
                            sim_max = sim
                    if sent in sents_i:
                        sim_j_ol += sim_max
                    else:
                        sim_j_nol += sim_max
                        cont_j_nol_sents += 1
                sim_i = sim_i_ol / cont_ol_sents
                if cont_i_nol_sents != 0:
                    sim_i = sim_i + (1 - sim_i) * sim_i_nol / float(cont_i_nol_sents)
                sim_j = sim_j_ol / cont_ol_sents
                if cont_j_nol_sents != 0:
                    sim_j = sim_j + (1 - sim_j) * sim_j_nol / float(cont_j_nol_sents)
                if sim_i > 0.99 and sim_j > 0.99:
                    if len(sents_j) > len(sents_i):
                        if sim_j > higher_sim:
                            ind_max = i + j
                            higher_sim = sim_j
                    else:
                        if sim_i > higher_sim:
                            ind_max = i
                            higher_sim = sim_i
                elif sim_j > sim_i:
                    if sim_j > higher_sim:
                        ind_max = i + j
                        higher_sim = sim_j
                    elif sim_i > higher_sim:
                        ind_max = i
                        higher_sim = sim_i
            if flag == 0:
                res_plags.append(plags[ind_max])
                res_psr.append(psr[ind_max])
            elif ind_max != i:
                del res_plags[-1]
                del res_psr[-1]
                res_plags.append(plags[ind_max])
                res_psr.append(psr[ind_max])
            i = i + cont_ol
            flag = 1
    return res_plags, res_psr


def remove_small_plags(plags, psr, src_offsets, susp_offsets, th):
    """
    DESCRIPTION: Remove the plagiarism cases that have less tha th characters either in the source or suspicios
    fragments
    INPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Have the plagiarism cases represented by min
    and max sentence index in suspicious and source document respectively
           psr <list of list of tuples (int, int)> - Contains the clusters
           src_offsets <list of tuples (int, int)> - Contain the char offset and length of each source document sentence
           susp_offsets <list of tuples (int, int)> - Contain the char offset and length of each suspicious document
           sentence
    OUTPUT: res_plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases without
    short cases
            res_psr <list of list of tuples (int, int)> - Contains the clusters without short cases
    """
    res_plags = []
    res_psr = []
    for i in range(len(plags)):
        arg1 = (susp_offsets[plags[i][0][0]][0], susp_offsets[plags[i][0][1]][0] + susp_offsets[plags[i][0][1]][1])
        arg2 = (src_offsets[plags[i][1][0]][0], src_offsets[plags[i][1][1]][0] + src_offsets[plags[i][1][1]][1])
        if arg1[1] - arg1[0] >= th and arg2[1] - arg2[0] >= th:
            res_plags.append(plags[i])
            res_psr.append(psr[i])
    return res_plags, res_psr


def word_span_tokenizer(text):
    """
    DESCRIPTION: Tokenize a text in words
    INPUT: text <string> - Text to be tokenized
    OUTPUT: words <list> - List of words from text
            offsets <list of tuple (int, int)> - Initial and final position of each word
    """
    words = []
    offsets = []
    sent_detector = nltk.data.load('tokenizers/punkt/english.pickle')
    word_detector = nltk.TreebankWordTokenizer()
    punctuation = string.punctuation
    for span in sent_detector.span_tokenize(text):
        sent = text[span[0]:span[1]].lower()
        sent_words = []
        for token in word_detector.tokenize(sent):
            for char in token:
                if char not in punctuation:
                    sent_words.append(token)
                    break
        idx = 0
        for word in sent_words:
            words.append(word)
            pos = sent[idx:].find(word)
            offsets.append([span[0] + idx + pos, idx + span[0] + pos + len(word)])  # (Initial position, Final position)
            if idx == 0:  # changing first word offset
                offsets[-1][0] = span[0]
            idx = idx + pos + len(word)
        if len(words) > 0:  # Changing last word offset
            offsets[-1][1] = span[1]
    return words, offsets


# Using Dynamic programming to find all elements greater then th (not only biggest element)
def longest_common_substring_all(s1, s1_off, s2, s2_off, th):
    """
    DESCRIPTION: Find the common subtrings using dynamic programming
    INPUT: s1 <list> - List of words from text 1
           s1_off <list of tuple (int, int)> - List of offsets of text1
           s2 <list> - List of words from text 2
           s2_off <list of tuple (int, int)> - List of offsets of text2
           th <int> - Threshold in characters of shortest common substring allowed
    OUTPUT: res <list tuples (int, int, int, int)> - Common subtring correspondence in text1 and text2 represented as
    char offsets (t1_init, t1_end, t2_init, t2_end)
    """
    m = [[0] * (1 + len(s2)) for i in range(1 + len(s1))]
    res = []
    longest, x_longest, y_longest = 0, 0, 0
    for x in range(1, 1 + len(s1)):
        for y in range(1, 1 + len(s2)):
            if s1[x - 1] == s2[y - 1]:
                m[x][y] = m[x - 1][y - 1] + 1
                if m[x][y] > longest:
                    longest = m[x][y]
            else:
                m[x][y] = 0
                if m[x - 1][y - 1] != 0:
                    len_plag = s1_off[x - 2][1] - s1_off[x - 1 - m[x - 1][y - 1]][0]
                    if len_plag > th:
                        res.append((s1_off[x - 1 - m[x - 1][y - 1]][0], s1_off[x - 2][1],
                                    s2_off[y - 1 - m[x - 1][y - 1]][0], s2_off[y - 2][1]))
        if m[x][y] != 0:  # Last column
            len_plag = s1_off[x - 1][1] - s1_off[x - m[x][y]][0]
            if len_plag > th:
                res.append((s1_off[x - m[x][y]][0], s1_off[x - 1][1], s2_off[y - m[x][y]][0], s2_off[y - 1][1]))
    for y in range(1, len(s2)):  # Last row
        if m[-1][y] != 0:
            len_plag = s1_off[-1][1] - s1_off[len(s1_off) - m[-1][y]][0]
            if len_plag > th:
                res.append(
                    (s1_off[len(s1_off) - m[-1][y]][0], s1_off[- 1][1], s2_off[y - m[-1][y]][0], s2_off[y - 1][1]))
    return res


def common_substring_pro_all(str1, str2, th_acc):
    """
    DESCRIPTION: Find the common substrings longer than some threshold
    INPUT: str1 <list> - Text 1
           str2 <list> - Text 2
           th_acc <int> - Threshold in characters of shortest common substring allowed
    OUTPUT: res <list tuples (int, int, int, int)> - Common subtring correspondence in text1 and text2 represented as
    char offsets (t1_init, t1_end, t2_init, t2_end)
    """
    X, X_off = word_span_tokenizer(str1)
    Y, Y_off = word_span_tokenizer(str2)
    res = longest_common_substring_all(X, X_off, Y, Y_off, th_acc)
    return res


def verbatim_det_lcs_all(plags, psr, susp_text, src_text, susp_offsets, src_offsets, th_shortest):
    """
    DESCRIPTION: Uses longest common substring algorithm to classify a pair of documents being compared as verbatim
    plagarism candidate (the pair of documents), and removing the none verbatim cases if positive
    INPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Have the plagiarism cases represented by min
    and max sentence index in suspicious and source document respectively
           psr <list of list of tuples (int, int)> - Contains the clusters
           susp_text <string> - Suspicios document text
           src_text <string> - Source document text
           susp_offsets <list of tuples (int, int)> - Contain the char offset and length of each suspicious document
           sentence
           src_offsets <list of tuples (int, int)> - Contain the char offset and length of each source document sentence
           th_shortest <int> - Threshold in characters of shortest common substring allowed
    OUTPUT: res_plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases as common
    substrings or the same as the arguments depending on type_plag
            res_psr <list of list of tuples (int, int)> - Contains the clusters with seeds present in the common
            substrings, or the same as the arguments depending on type_plag
            type_plag <0 or 1> - 1: verbatim plagiarism case    0: Other plagiarism case
            res_long_frag <list> - Contains the lengths of common substrings
    """
    # plags[[(susp_ini, susp_end), (src_ini, src_end)], ...]
    res_plags = []
    res_psr = []
    res_long_frag = []
    i = 0
    type_plag = 0  # 0: Unknown, 1: no-obfuscation
    while i < len(plags):  # For each plagiarism case
        # print('Case',i)
        # print('Plag case', plags[i])
        # print('Seeds', psr[i])
        # sentences in seeds and those not in seeds
        res2 = common_substring_pro_all(susp_text[susp_offsets[plags[i][0][0]][0]: susp_offsets[plags[i][0][1]][0] +
                                                                                   susp_offsets[plags[i][0][1]][1]],
                                        src_text[src_offsets[plags[i][1][0]][0]: src_offsets[plags[i][1][1]][0] +
                                                                                 src_offsets[plags[i][1][1]][1]],
                                        th_shortest)
        res = []
        # Remove overlapping
        for tup_i in res2:
            flag = 0
            for tup_j in res2:
                if tup_i != tup_j and tup_i[2] >= tup_j[2] and tup_i[3] <= tup_j[3]:
                    flag = 1
                    break
            if flag == 0:
                res.append(tup_i)

        if len(res) > 0:
            if type_plag == 1:
                # Removing seeds with lcs shorter than th_shortest
                for sub_case in res:
                    res_plags.append(
                        [(susp_offsets[plags[i][0][0]][0] + sub_case[0], susp_offsets[plags[i][0][0]][0] + sub_case[1]),
                         (src_offsets[plags[i][1][0]][0] + sub_case[2], src_offsets[plags[i][1][0]][0] + sub_case[3])])
                    res_psr.append(psr[i])
                    res_long_frag.append(max([sub_case[1] - sub_case[0], sub_case[3] - sub_case[2]]))
            else:
                # Type 02-no-obfuscation detected. Starting over! Removing previously added cases!
                type_plag = 1
                res_plags = []
                res_psr = []
                res_long_frag = []
                for sub_case in res:
                    res_plags.append(
                        [(susp_offsets[plags[i][0][0]][0] + sub_case[0], susp_offsets[plags[i][0][0]][0] + sub_case[1]),
                         (src_offsets[plags[i][1][0]][0] + sub_case[2], src_offsets[plags[i][1][0]][0] + sub_case[3])])
                    res_psr.append(psr[i])
                    res_long_frag.append(max([sub_case[1] - sub_case[0], sub_case[3] - sub_case[2]]))
                # i = -1
        else:
            if type_plag != 1:
                res_plags.append(plags[i])
                res_psr.append(psr[i])
                res_long_frag.append(-1)
        i += 1
    return res_plags, res_psr, type_plag, res_long_frag


"""
MAIN CLASS
"""


class SGSPLAG:
    def __init__(self, susp_text, src_text, config):
        self.__dict__.update(config)

        self.susp_text = susp_text
        self.src_text = src_text
        self.src_voc = {}
        self.susp_voc = {}
        self.src_offsets = []
        self.susp_offsets = []
        self.src_bow = {}
        self.susp_bow = {}
        self.detections = None

    def process(self, preprocessor, sent_matcher):
        """
        DESCRIPTION: Process the plagiarism pipeline
        INPUT: self <SGSPLAG object>
        OUTPUT: type_plag <int> - Verbatim plagarism flag
                summary_flag <int> - Summary plagarism flag
        """
        preprocessor.preprocess(self)
        self.detections, type_plag, summary_flag = self.compare(sent_matcher)
        return type_plag, summary_flag

    def extension(self, ps):
        """
        DESCRIPTION: Adding two vectors
        INPUT: self <SGSPLAG object>
               ps <list of tuple (int, int, float, float)> - Seeds
        OUTPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases after
        validation
                psr <list of list of tuples (int, int)> - Contains the clusters after validation
                sim_frag <list of floats> - Stores the cosine similarity between source and suspicios fragments in
                the plagiarism cases after validation
        """
        psr = clustering(ps, self.src_offsets, self.susp_offsets, self.src_gap, self.susp_gap, self.src_size,
                         self.susp_size, 0, 0)
        plags = []
        for psr_i in psr:
            plags.append([(min([x[0] for x in psr_i]), max([x[0] for x in psr_i])),
                          (min([x[1] for x in psr_i]), max([x[1] for x in psr_i]))])
        temp_res = validation(plags, psr, self.src_offsets, self.susp_offsets, self.src_bow, self.susp_bow,
                              self.src_gap, self.src_gap_least, self.susp_gap, self.susp_gap_least, self.src_size,
                              self.susp_size, self.th3)
        if len(temp_res) == 0:
            plags, psr, sim_frag = [], [], []
        else:
            plags, psr, sim_frag = temp_res[0], temp_res[1], temp_res[2]
        return plags, psr, sim_frag

    def filtering(self, plags, psr):
        """
        DESCRIPTION: Filter the plagiarism cases by removing overlapping and short cases
        INPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases after
        validation
               psr <list of list of tuples (int, int)> - Contains the clusters after validation
        OUTPUT: plags <list of list of two tuples [(int, int), (int, int)]> - Contains the plagiarism cases. Also
        modify psr.
        """
        plags, psr = remove_overlap3(plags, psr, self.src_bow, self.susp_bow)
        plags, psr = remove_small_plags(plags, psr, self.src_offsets, self.susp_offsets, self.min_plaglen)

        return plags

    def compare(self, sent_matcher):
        """
        DESCRIPTION: Test a suspicious document for near-duplicate plagiarism with regards to a source document and
        return a feature list depending on the type_plag and summary_flag flags.
        INPUT: self <SGSPLAG object>
        OUTPUT: detections <list> - Representation of plagairism cases before writing the xml file with require PAN
        format
                type_plag <int> - Verbatim flag
                summary_flag <int> - Summary flag
        """
        detections = []
        ps = sent_matcher.seeding(self)
        plags, psr, sim_frag = self.extension(ps)
        plags = self.filtering(plags, psr)
        if self.verbatim != 0:
            plags_verbatim, psr_verbatim, type_plag, long_frag = verbatim_det_lcs_all(plags, psr, self.susp_text,
                                                                                      self.src_text, self.susp_offsets,
                                                                                      self.src_offsets,
                                                                                      self.verbatim_minlen)
        else:
            type_plag = 0

        ####META-CLASSIFIER####
        if self.summary != 0:
            self.src_gap = self.src_gap_summary
            self.susp_gap = self.susp_gap_summary
            plags2, psr2, sim_frag = self.extension(ps)
            plags2 = self.filtering(plags2, psr2)
        summary_flag = 0
        if type_plag == 0:
            sum_src = 0
            sum_susp = 0
            if self.summary != 0:
                for plag in plags2:
                    arg1 = (self.susp_offsets[plag[0][0]][0],
                            self.susp_offsets[plag[0][1]][0] + self.susp_offsets[plag[0][1]][1])
                    arg2 = (
                        self.src_offsets[plag[1][0]][0],
                        self.src_offsets[plag[1][1]][0] + self.src_offsets[plag[1][1]][1])
                    sum_susp = sum_susp + (arg1[1] - arg1[0])
                    sum_src = sum_src + (arg2[1] - arg2[0])
            if sum_src != 0 and sum_src >= 3 * sum_susp:  # Summary heuristic
                summary_flag = 1
                for plag in plags2:
                    arg1 = (self.susp_offsets[plag[0][0]][0],
                            self.susp_offsets[plag[0][1]][0] + self.susp_offsets[plag[0][1]][1])
                    arg2 = (
                        self.src_offsets[plag[1][0]][0],
                        self.src_offsets[plag[1][1]][0] + self.src_offsets[plag[1][1]][1])
                    detections.append([arg1, arg2])
            else:
                for plag in plags:
                    arg1 = (self.susp_offsets[plag[0][0]][0],
                            self.susp_offsets[plag[0][1]][0] + self.susp_offsets[plag[0][1]][1])
                    arg2 = (
                        self.src_offsets[plag[1][0]][0],
                        self.src_offsets[plag[1][1]][0] + self.src_offsets[plag[1][1]][1])
                    detections.append([arg1, arg2])
        else:
            for plag in plags_verbatim:
                arg1 = plag[0][0], plag[0][1]
                arg2 = plag[1][0], plag[1][1]
                detections.append([arg1, arg2])
        return detections, type_plag, summary_flag


@dataclass(frozen=True)
class Section:
    doc: Document
    offset: int
    length: int


@dataclass(frozen=True)
class Match:
    sec1: Section
    sec2: Section


class DifferentDocumentPairError(Exception):
    pass


class InvalidConfigError(Exception):
    pass


class DocumentPairMatches:
    def __init__(self):
        self._matches: list[Match] = []

    def add(self, match: Match):
        if self._matches:
            doc_pair1, doc_pair2 = (self._matches[-1].sec1.doc, self._matches[-1].sec2.doc), \
                                   (match.sec1.doc, match.sec2.doc)
            if not doc_pair1 == doc_pair2:
                raise DifferentDocumentPairError(f'Only matches of document pair {doc_pair1} can be added.')
        self._matches.append(match)

    @property
    def doc1(self) -> Document:
        if not self.empty():
            return self._matches[0].sec1.doc

    @property
    def doc2(self) -> Document:
        if not self.empty():
            return self._matches[0].sec2.doc

    def list(self) -> [Match]:
        return self._matches

    def empty(self) -> bool:
        return not self._matches


def find_matches(docs: list[Document], lang: str, config: dict) -> list[DocumentPairMatches]:
    matches = []
    try:
        preprocessor = Preprocessor(lang, config['min_sent_len'], config['rem_stop_words'])
        preprocessor.preprocess_new(docs)
        for doc1, doc2 in combinations(docs, 2):
            doc_pair_matches = DocumentPairMatches()
            sgsplag_obj = SGSPLAG(doc1.text, doc2.text, config)

            preprocessor.set_docs(doc1, doc2)
            sgsplag_obj.process(preprocessor, SentenceMatcher(config['th1'], config['th2']))

            for det in sgsplag_obj.detections:
                match = Match(Section(doc1, det[0][0], det[0][1] - det[0][0]),
                              Section(doc2, det[1][0], det[1][1] - det[1][0]))
                doc_pair_matches.add(match)
            if not doc_pair_matches.empty():
                matches.append(doc_pair_matches)
        return matches
    except (KeyError, AttributeError) as e:
        raise InvalidConfigError('The given config seems to be invalid.') from e



import sys

import click

from plagdef import bootstrap
from plagdef.model.reporting import generate_text_report

# Core functions of PlagDef
find_matches = bootstrap.find_matches()
write_xml_reports = bootstrap.write_xml_reports()


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.argument('docdir', type=click.Path(exists=True))
@click.option('lang', '--lang', '-l', required=True, help='Bibliographical language code of the documents\' language.')
@click.option('rec', '--recursive', '-R', is_flag=True,
              help='Recursively search for documents in subdirectories.')
@click.option('xmldir', '--xml', '-x', type=click.Path(), help='Output directory for XML reports.')
def cli(docdir: click.Path, lang: str, xmldir: click.Path, rec: bool):
    """
    \b
    PlagDef supports plagiarism detection for student assignments.
    It must be provided with a directory <DOCDIR> containing at least two documents and their language.
    """
    matches = find_matches(docdir, lang, recursive=rec)
    click.echo(f'Found {len(matches) if len(matches) else "no"} suspicious document pair'
               f'{"s" if len(matches) > 1 else ""}.\n')
    if matches:
        if xmldir:
            write_xml_reports(matches, xmldir)
            click.echo(f'Successfully wrote XML reports to {xmldir}.')
        else:
            text_report = generate_text_report(matches)
            click.echo(text_report)
    sys.exit(0)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
def gui():
    """
    \b
    PlagDef supports plagiarism detection for student assignments.
    The GUI for this tool is based on the Qt 6 framework and works on all platforms.
    """
    from plagdef.gui.main import MyQtApp
    app = MyQtApp(find_matches)
    app.window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    args = sys.argv
    if len(args) > 1 and args[1] == 'gui':
        del args[1]
        gui()
    else:
        cli()



from __future__ import annotations

from ast import literal_eval
from configparser import ConfigParser
from itertools import islice
from pathlib import Path

import chardet

from plagdef.model.legacy.algorithm import Document


class DocumentFileRepository:
    def __init__(self, dir_path: Path, lang: str, recursive=False):
        self.lang = lang
        self._documents = []
        if not dir_path.is_dir():
            raise NotADirectoryError(f'The given path {dir_path} does not point to an existing directory!')
        if not any(dir_path.iterdir()) or not next(islice(dir_path.iterdir(), 1, None), None):
            raise NoDocumentFilePairFoundError(f'The directory {dir_path} must contain at least two documents.')
        if recursive:
            doc_files = [file_path for file_path in dir_path.rglob('*') if file_path.is_file()]
        else:
            doc_files = [file_path for file_path in dir_path.iterdir() if file_path.is_file()]
        try:
            for file in doc_files:
                raw_data = file.read_bytes()
                enc = chardet.detect(raw_data)['encoding']
                self._documents.append(Document(file.stem, file.read_text(encoding=enc)))
        except UnicodeDecodeError as e:
            raise UnsupportedFileFormatError(f'The directory {dir_path} contains files in an unsupported format.') \
                from e

    def list(self) -> [Document]:
        return self._documents


class DocumentPairReportFileRepository:
    def __init__(self, out_path: Path):
        if not out_path.is_dir():
            raise NotADirectoryError(f'The given path {out_path} does not point to an existing directory!')
        self._out_path = out_path

    def add(self, doc_pair_report):
        file_name = Path(f'{doc_pair_report.doc1.name}-{doc_pair_report.doc2.name}.{doc_pair_report.format}')
        file_path = self._out_path / file_name
        with file_path.open('w', encoding='utf-8') as f:
            f.write(doc_pair_report.content)


class ConfigFileRepository:
    def __init__(self, config_path: Path):
        if not config_path.is_file():
            raise FileNotFoundError(f'The given path {config_path} does not point to an existing file!')
        if not config_path.suffix == '.ini':
            raise UnsupportedFileFormatError(f'The config file format must be INI.')
        self.config_path = config_path

    def get(self) -> dict:
        parser = ConfigParser()
        parser.read(self.config_path)
        config = {}
        for section in parser.sections():
            typed_config = [(key, literal_eval(val)) for key, val in parser.items(section)]
            config.update(dict(typed_config))
        return config


class NoDocumentFilePairFoundError(Exception):
    pass


class UnsupportedFileFormatError(Exception):
    pass



from __future__ import annotations

import sys
from pathlib import Path

import pkg_resources
from PySide6.QtCore import QFile, QIODevice, Qt
from PySide6.QtGui import QCursor, QMovie
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QButtonGroup, QMainWindow, QFileDialog, QDialog

from plagdef.gui.model import ResultsTableModel

UI_FILES = {
    'main_window': pkg_resources.resource_filename(__name__, 'ui/main_window.ui'),
    'home_widget': pkg_resources.resource_filename(__name__, 'ui/home_widget.ui'),
    'loading_widget': pkg_resources.resource_filename(__name__, 'ui/loading_widget.ui'),
    'error_widget': pkg_resources.resource_filename(__name__, 'ui/error_widget.ui'),
    'no_results_widget': pkg_resources.resource_filename(__name__, 'ui/no_results_widget.ui'),
    'results_widget': pkg_resources.resource_filename(__name__, 'ui/results_widget.ui'),
}


class MainWindow:
    def __init__(self, views: list[View]):
        self._window = _load_ui_file(Path(UI_FILES['main_window']))
        self._views = views
        self._configure()

    def _configure(self):
        for view in self._views:
            self._window.stacked_widget.addWidget(view.widget)

    def switch_to(self, view_cls: type, data=None):
        idx = self._window.stacked_widget.currentIndex()
        for view in self._views:
            if type(view) == view_cls:
                self._views[idx].on_destroy()
                self._window.stacked_widget.setCurrentWidget(view.widget)
                view.on_init(data)

    def show(self):
        self._window.show()


class View:
    def on_init(self, data=None):
        pass

    def on_destroy(self):
        pass


class HomeView(View):
    def __init__(self):
        self.lang = None
        self.recursive = False
        self.widget = _load_ui_file(Path(UI_FILES['home_widget']))
        self._configure()

    def _configure(self):
        self.widget.lang_button_group = QButtonGroup()
        self.widget.lang_button_group.addButton(self.widget.eng_radio)
        self.widget.lang_button_group.addButton(self.widget.ger_radio)
        self.widget.open_folder_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.detect_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.folder_name_label.setVisible(False)
        self.widget.recursive_check_box.setVisible(False)
        self.widget.remove_folder_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.remove_folder_button.setVisible(False)
        self.widget.sel_lang_label.setVisible(False)
        self.widget.change_lang_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.change_lang_button.setVisible(False)

    def register_for_signals(self, open_folder=None, remove_folder=None, select_lang=None,
                             reset_lang_sel=None, detect=None):
        self.widget.open_folder_button.clicked.connect(lambda: open_folder())
        self.widget.remove_folder_button.clicked.connect(lambda: remove_folder())
        self.widget.lang_button_group.buttonClicked.connect(lambda: select_lang())
        self.widget.change_lang_button.clicked.connect(lambda: reset_lang_sel())
        self.widget.detect_button.clicked.connect(lambda: detect())

    def folder_selected(self, folder_name: str):
        self.widget.folder_name_label.setText(
            f'<html><head/><body><p align="center"><span style="font-size:12pt; color:#ffffff;"> '
            f'{folder_name}</span></p></body></html>')
        self.widget.open_folder_button.setVisible(False)
        self.widget.folder_name_label.setVisible(True)
        self.widget.recursive_check_box.setVisible(True)
        self.widget.remove_folder_button.setVisible(True)
        [radio.setEnabled(True) for radio in self.widget.lang_button_group.buttons()]

    def reset_folder_selection(self):
        self.recursive = False
        self.widget.recursive_check_box.setChecked(False)
        self.widget.open_folder_button.setVisible(True)
        self.widget.folder_name_label.setVisible(False)
        self.widget.recursive_check_box.setVisible(False)
        self.widget.remove_folder_button.setVisible(False)
        [radio.setEnabled(False) for radio in self.widget.lang_button_group.buttons()]

    def language_selected(self):
        self.recursive = self.widget.recursive_check_box.isChecked()
        lang_long = self.widget.lang_button_group.checkedButton().text()
        self.lang = lang_long[:3].lower()
        self.widget.sel_lang_label.setText(
            f'<html><head/><body><p align="center"><span style="font-size:12pt; color:#ffffff;"> '
            f'{lang_long}</span></p></body></html>')
        self.widget.recursive_check_box.setEnabled(False)
        self.widget.remove_folder_button.setVisible(False)
        [radio.setVisible(False) for radio in self.widget.lang_button_group.buttons()]
        self.widget.sel_lang_label.setVisible(True)
        self.widget.change_lang_button.setVisible(True)
        self.widget.detect_button.setEnabled(True)

    def reset_language_selection(self):
        self.lang = None
        self.widget.lang_button_group.setExclusive(False)
        self.widget.lang_button_group.checkedButton().setChecked(False)
        self.widget.lang_button_group.setExclusive(True)
        self.widget.detect_button.setEnabled(False)
        self.widget.sel_lang_label.setVisible(False)
        self.widget.change_lang_button.setVisible(False)
        [radio.setVisible(True) for radio in self.widget.lang_button_group.buttons()]
        self.widget.remove_folder_button.setVisible(True)
        self.widget.recursive_check_box.setEnabled(True)

    def on_destroy(self):
        self.reset_language_selection()
        [radio.setEnabled(False) for radio in self.widget.lang_button_group.buttons()]
        self.reset_folder_selection()


class FileDialog(QFileDialog):
    def __init__(self):
        super().__init__()
        self.setFileMode(QFileDialog.Directory)
        self.selected_dir = None

    def open(self) -> bool:
        if self.exec_() == QDialog.Accepted:
            self.selected_dir = self.selectedFiles()[0]
            return True


class LoadingView(View):
    def __init__(self):
        self._movie = QMovie(':/loading.gif')
        self.widget = _load_ui_file(Path(UI_FILES['loading_widget']))
        self._configure()

    def _configure(self):
        self.widget.loading_movie_label.setMovie(self._movie)

    def on_init(self, data=None):
        self._movie.start()

    def on_destroy(self):
        self._movie.stop()


class NoResultsView(View):
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['no_results_widget']))
        self._configure()

    def _configure(self):
        self.widget.again_button_no_res.setCursor(QCursor(Qt.PointingHandCursor))

    def register_for_signals(self, again=None):
        self.widget.again_button_no_res.clicked.connect(lambda: again())


class ErrorView(View):
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['error_widget']))
        self._configure()

    def _configure(self):
        self.widget.again_button_err.setCursor(QCursor(Qt.PointingHandCursor))

    def on_init(self, data=None):
        self.widget.error_msg_label.setText(
            f'<html><head/><body><p align="center"><span style=" font-size:12pt; color:#ffffff;"> '
            f'{data}</span></p></body></html>')
        self.widget.error_msg_label.setWordWrap(True)

    def register_for_signals(self, again=None):
        self.widget.again_button_err.clicked.connect(lambda: again())


class ResultView(View):
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['results_widget']))
        self._configure()

    def _configure(self):
        self.widget.again_button_res.setCursor(QCursor(Qt.PointingHandCursor))

    def on_init(self, data=None):
        model = ResultsTableModel(data)
        self.widget.results_table.setModel(model)
        self.widget.results_table.resizeRowsToContents()
        self.widget.results_table.resizeColumnsToContents()

    def register_for_signals(self, again=None):
        self.widget.again_button_res.clicked.connect(lambda: again())


def _load_ui_file(path: Path) -> QMainWindow:
    main_ui = QFile(str(path))
    if not main_ui.open(QIODevice.ReadOnly):
        print(f"Cannot open {path}: {main_ui.errorString()}")
        sys.exit(-1)
    loader = QUiLoader()
    window = loader.load(main_ui)
    main_ui.close()
    if not window:
        print(loader.errorString())
        sys.exit(-1)
    return window



import plagdef.gui.main as main
from plagdef.gui.views import HomeView, LoadingView, NoResultsView, ErrorView, ResultView, \
    FileDialog


class HomeController:
    def __init__(self):
        self.view = HomeView()
        self.file_dialog = FileDialog()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(open_folder=self._on_open_folder, remove_folder=self._on_remove_folder,
                                       select_lang=self._on_select_lang, reset_lang_sel=self._on_reset_lang,
                                       detect=self._on_detect)

    def _on_open_folder(self):
        if self.file_dialog.open():
            folder_name = self.file_dialog.selected_dir[self.file_dialog.selected_dir.rfind("/"):]
            self.view.folder_selected(folder_name)

    def _on_remove_folder(self):
        self.view.reset_folder_selection()

    def _on_select_lang(self):
        self.view.language_selected()

    def _on_reset_lang(self):
        self.view.reset_language_selection()

    def _on_detect(self):
        main.app.find_matches(self.file_dialog.selected_dir, self.view.lang, recursive=self.view.recursive)
        main.app.window.switch_to(LoadingView)


class LoadingController:
    def __init__(self):
        self.view = LoadingView()


class NoResultsController:
    def __init__(self):
        self.view = NoResultsView()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(self._on_again)

    def _on_again(self):
        main.app.window.switch_to(HomeView)


class ErrorController:
    def __init__(self):
        self.view = ErrorView()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(self._on_again)

    def _on_again(self):
        main.app.window.switch_to(HomeView)


class ResultController:
    def __init__(self):
        self.view = ResultView()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(self._on_again)

    def _on_again(self):
        main.app.window.switch_to(HomeView)



from dataclasses import dataclass

from numpy import dot
from numpy.linalg import norm

from plagdef.model.preprocessing import Sentence


@dataclass(frozen=True)
class SentenceMatch:
    sent1: Sentence
    sent2: Sentence
    cos_sim: float
    dice_sim: float


class SentenceMatcher:
    def __init__(self, min_cos_sim: float, min_dice_sim: float):
        self._min_cos_sim = min_cos_sim
        self._min_dice_sim = min_dice_sim

    def match(self, sent1: Sentence, sent2: Sentence):
        cos_sim = self._cos_sim(sent1, sent2)
        dice_sim = self._dice_sim(sent1, sent2)
        if cos_sim > self._min_cos_sim and dice_sim > self._min_dice_sim:
            return SentenceMatch(sent1, sent2, cos_sim, dice_sim)

    def _cos_sim(self, sent1: Sentence, sent2: Sentence):
        """
        Compute the cosine similarity cos-sim = sum_i=1_n{a_i * b_i} / sqrt{sum_i=1_n{(a_i)^2}} * sqrt{sum_i=1_n{(
        b_i)^2}}
        """
        aligned_vecs = [(sent1.bow_tf_isf[lemma], sent2.bow_tf_isf[lemma]) for lemma in sent1.bow_tf_isf if
                        lemma in sent2.bow_tf_isf]
        a, b = [v[0] for v in aligned_vecs], [v[1] for v in aligned_vecs]
        return dot(a, b) / (norm(list(sent1.bow_tf_isf.values())) * norm(list(sent2.bow_tf_isf.values())))

    def _dice_sim(self, sent1: Sentence, sent2: Sentence):
        """
        Compute the dice similarity dice-coeff = 2 * n_com / (n_x + n_y), n_t being the number of distinct common
        lemmas in both sents, n_1 the number of lemmas in sent1 and n_2 the number of lemmas in sent2
        """
        n_com = len(set(sent1.bow_tf_isf.keys()).intersection(sent2.bow_tf_isf.keys()))
        return 2 * n_com / float(len(sent1.bow_tf_isf) + len(sent2.bow_tf_isf))

    # TODO: Only for backwards compatibility
    def seeding(self, legacy_obj):
        sent_matches = []
        for doc1_sent_num in range(len(legacy_obj.susp_bow)):
            for doc2_sent_num in range(len(legacy_obj.src_bow)):
                match = self.match(Sentence(None, -1, -1, None, legacy_obj.susp_bow[doc1_sent_num]),
                                   Sentence(None, -1, -1, None, legacy_obj.src_bow[doc2_sent_num]))
                if match:
                    sent_matches.append((doc1_sent_num, doc2_sent_num, match.cos_sim, match.dice_sim))
        return sent_matches



from dataclasses import dataclass

from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtGui import QColor

from plagdef.model.legacy import algorithm


@dataclass(frozen=True)
class ResultRow:
    doc1: str
    doc1_offset: int
    doc1_length: int
    doc2: str
    doc2_offset: int
    doc2_length: int


class ResultsTableModel(QAbstractTableModel):
    def __init__(self, matches: list[algorithm.DocumentPairMatches]):
        super().__init__()
        self._rows = []
        for doc_pair_matches in matches:
            for match in doc_pair_matches.list():
                self._rows.append(ResultRow(doc_pair_matches.doc1.name, match.sec1.offset, match.sec1.length,
                                            doc_pair_matches.doc2.name, match.sec2.offset, match.sec2.length))

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole):
        headers = ['Document 1', 'Offset', 'Length', 'Document 2', 'Offset', 'Length']
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return headers[section]

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 6

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._rows)

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        match = self._rows[index.row()]
        match_attributes = [match.doc1, match.doc1_offset, match.doc1_length,
                            match.doc2, match.doc2_offset, match.doc2_length]
        if role == Qt.DisplayRole:
            return match_attributes[index.column()]
        elif role == Qt.ForegroundRole:
            return QColor(255, 255, 255)



import math


class LegacySentenceMatcher:
    def seeding(self, legacy_obj):
        """
        DESCRIPTION: Creates the seeds from pair of sentece similarity using dice and cosine similarity
        INPUT: self <SGSPLAG object>
        OUTPUT: ps <list of tuple (int, int, float, float)> - Seeds
        """
        ps = []
        for c in range(len(legacy_obj.susp_bow)):
            for r in range(len(legacy_obj.src_bow)):
                v1 = self._cosine_measure(legacy_obj.susp_bow[c], legacy_obj.src_bow[r])
                v2 = self._dice_coeff(legacy_obj.susp_bow[c], legacy_obj.src_bow[r])
                if v1 > legacy_obj.th1 and v2 > legacy_obj.th2:
                    ps.append((c, r, v1, v2))
        return ps

    def _cosine_measure(self, d1, d2):
        """
        DESCRIPTION: Compute the cosine measure (cosine of the angle between two vectors) in sparse (dictionary)
        representation
        INPUT: d1 <dictionary> - Sparse vector 1
               d2 <dictionary> - Sparse vector 2
        OUTPUT: Cosine measure
        """
        dot_prod = 0.0
        det = self._eucl_norm(d1) * self._eucl_norm(d2)
        if det == 0:
            return 0
        for word in d1.keys():
            if word in d2:
                dot_prod += d1[word] * d2[word]
        return dot_prod / det

    def _dice_coeff(self, d1, d2):
        """
        DESCRIPTION: Compute the dice coefficient in sparse (dictionary) representation
        INPUT: d1 <dictionary> - Sparse vector 1
               d2 <dictionary> - Sparse vector 2
        OUTPUT: Dice coefficient
        """
        if len(d1) + len(d2) == 0:
            return 0
        intj = 0
        for i in d1.keys():
            if i in d2:
                intj += 1
        return 2 * intj / float(len(d1) + len(d2))

    def _eucl_norm(self, d1):
        """
        DESCRIPTION: Compute the Euclidean norm of a sparse vector
        INPUT: d1 <dictionary> - sparse vector representation
        OUTPUT: Norm of the sparse vector d1
        """
        norm = 0.0
        for val in d1.values():
            norm += float(val * val)
        return math.sqrt(norm)



from __future__ import annotations

import sys
import traceback
from collections import Callable

from PySide6.QtCore import QThreadPool, QRunnable, Slot, QObject, Signal
from PySide6.QtWidgets import QApplication
from click import UsageError

# noinspection PyUnresolvedReferences
import plagdef.gui.resources
from plagdef.gui.controllers import HomeController, LoadingController, ErrorController, NoResultsController, \
    ResultController
from plagdef.gui.views import MainWindow, ResultView, NoResultsView, ErrorView
from plagdef.model.legacy.algorithm import DocumentPairMatches

app = None


class MyQtApp(QApplication):
    def __init__(self, find_matches: Callable):
        super().__init__()
        self._ctrls = [HomeController(), LoadingController(), ErrorController(), NoResultsController(),
                       ResultController()]
        self._views = [ctrl.view for ctrl in self._ctrls]
        self.window = MainWindow(self._views)
        self._find_matches = find_matches
        global app
        app = self

    def find_matches(self, doc_dir: str, lang: str, recursive=False):
        worker = Worker(self._find_matches, doc_dir, lang, recursive=recursive)
        worker.signals.result.connect(self._on_success)
        worker.signals.error.connect(self._on_error)
        pool = QThreadPool.globalInstance()
        pool.start(worker)

    def _on_success(self, matches: list[DocumentPairMatches]):
        if matches:
            self.window.switch_to(ResultView, matches)
        else:
            self.switch_to(NoResultsView)

    def _on_error(self, error: (type, Exception)):
        if error[0] == UsageError:
            self.window.switch_to(ErrorView, str(error[1]))
        else:
            self.window.switch_to(ErrorView, 'An error occurred. Please refer to the command line for more details.')
            raise error[1]


class Worker(QRunnable):
    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        try:
            result = self.fn(*self.args, **self.kwargs)
        except Exception:
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)


class WorkerSignals(QObject):
    error = Signal(tuple)
    result = Signal(object)



import copy
import math

import nltk
from Stemmer import Stemmer


class LegacyPreprocessor:
    def preprocess(self, legacy_obj):
        """
        DESCRIPTION: Preprocess the suspicious and source document
        INPUT: self <SGSPLAG object>
        OUTPUT: None. Gets bag of words with tf-idf, offsets and preprocess sentences
        """
        legacy_obj.src_bow = self._tokenize(legacy_obj.src_text, legacy_obj.src_voc,
                                            legacy_obj.src_offsets)
        self._join_small_sents(legacy_obj.src_bow, legacy_obj.src_offsets, legacy_obj.min_sent_len,
                               legacy_obj.src_voc)
        legacy_obj.susp_bow = self._tokenize(legacy_obj.susp_text, legacy_obj.susp_voc,
                                             legacy_obj.susp_offsets)
        self._join_small_sents(legacy_obj.susp_bow, legacy_obj.susp_offsets, legacy_obj.min_sent_len,
                               legacy_obj.susp_voc)

        self._tf_isf(legacy_obj.src_bow, legacy_obj.src_voc, legacy_obj.susp_bow,
                     legacy_obj.susp_voc)

    def _tokenize(self, text, voc={}, offsets=[], rem_sw=0):
        """
        DESCRIPTION: Tokenization and vectorization of sentences in a document
        INPUTS: text <string> - Text to be pre-processed
                voc <dictionary> - The keys are the types (vocabulary) in a document while the values are the sentence
                frequency
                offsets <2-tuple> - First value contains the offset character and the next contain the length in
                characters
                sents <list> - Sentences of the text without tokenization
                rem_sw <integer> - Option about treatment of stopwords (0): None stopword remove, (1): 50 more common
                stopwords removed, (other): All stopwords removed
        OUTPUT: sent_vects <list of dictionaries> - List of dictionaries representing each sentence vector. Sparce
        bag of
        words. Also modify sents, offsets and voc.
        NOTE: If char_preprocess() is used, you must use update_offsets() also
        """
        sents = []
        text = text.replace(chr(0), ' ')
        sent_detector = nltk.data.load('tokenizers/punkt/english.pickle')
        word_detector = nltk.TreebankWordTokenizer()
        stemmer = Stemmer('english')
        sent_spans = sent_detector.span_tokenize(text)
        sent_vects = []
        if rem_sw == 0:
            stopwords = []
        elif rem_sw == 1:
            stopwords = ['the', 'of', 'and', 'a', 'in', 'to', 'is', 'was', 'it', 'for', 'with', 'he', 'be', 'on', 'i',
                         'that', 'by', 'at', 'you', '\'s', 'are', 'not', 'his', 'this', 'from', 'but', 'had', 'which',
                         'she', 'they', 'or', 'an', 'were', 'we', 'their', 'been', 'has', 'have', 'will', 'would',
                         'her',
                         'n\'t', 'there', 'can', 'all', 'as', 'if', 'who', 'what', 'said']
        else:
            stopwords = ['i', 'me', 'my', 'myself', 'we', 'our', 'ours', 'ourselves', 'you', 'your', 'yours',
                         'yourself',
                         'yourselves', 'he', 'him', 'his', 'himself', 'she', 'her', 'hers', 'herself', 'it', 'its',
                         'itself', 'they', 'them', 'their', 'theirs', 'themselves', 'what', 'which', 'who', 'whom',
                         'this',
                         'that', 'these', 'those', 'am', 'is', 'are', 'was', 'were', 'be', 'been', 'being', 'have',
                         'has',
                         'had', 'having', 'do', 'does', 'did', 'doing', 'a', 'an', 'the', 'and', 'but', 'if', 'or',
                         'because', 'as', 'until', 'while', 'of', 'at', 'by', 'for', 'with', 'about', 'against',
                         'between',
                         'into', 'through', 'during', 'before', 'after', 'above', 'below', 'to', 'from', 'up', 'down',
                         'in',
                         'out', 'on', 'off', 'over', 'under', 'again', 'further', 'then', 'once', 'here', 'there',
                         'when',
                         'where', 'why', 'how', 'all', 'any', 'both', 'each', 'few', 'more', 'most', 'other', 'some',
                         'such', 'no', 'nor', 'not', 'only', 'own', 'same', 'so', 'than', 'too', 'very', '\'s', 'n\'t',
                         'can', 'will', 'just', 'don', 'should', 'now']
        for span in sent_spans:  # For each sentence
            sents.append(text[span[0]: span[1]].lower())
            sent_dic = {}
            for word in word_detector.tokenize(sents[-1]):  # for each word in the sentence
                if word[0].isalnum() and len(word) > 2:
                    if word not in stopwords:
                        word_pp = stemmer.stemWord(word)  # if word not in stopwords_all else word (costly)
                    else:
                        word_pp = word
                else:
                    continue
                if word_pp in sent_dic:
                    sent_dic[word_pp] += 1
                else:  # only one time for each sentence, increase voc[word] value
                    sent_dic[word_pp] = 1
                    if word_pp in voc:
                        voc[word_pp] += 1
                    else:
                        voc[word_pp] = 1
            if len(sent_dic) > 0:
                sent_vects.append(sent_dic)
                offsets.append([span[0], span[1] - span[0]])
        return sent_vects

    def _join_small_sents(self, list_dic, offsets, min_sentlen, voc):
        """
        DESCRIPTION: Remove or annex sentences with less than a certain amount of words (min_sentlen)
        INPUTS: list_dic <list of dictionaries> - List containing the vectors (dictionaries) of a document
                offsets <2-tuple> - First value contains the offset character and the next contain the length in
                characters
                min_sentlen <integer> - Minimum of words allowed in a sentence
                rssent <integer> - Action to perform (0): Annex small sentences, (1) Remove small sentences
                voc <dictionary> - The keys are the types (vocabulary) in a document while the values are the sentence
                frequency
        OUTPUT: No returned value. Modify the inputs list_dic, offsets and voc
        """
        i = 0
        range_i = len(list_dic) - 1
        while i < range_i:
            if sum(list_dic[i].values()) < min_sentlen:
                for k in list_dic[i].keys():
                    if k in list_dic[i + 1]:
                        voc[k] -= 1
                list_dic[i + 1] = self._sum_vect(list_dic[i + 1], list_dic[i])
                del list_dic[i]
                offsets[i + 1] = (offsets[i][0], offsets[i + 1][1] + offsets[i][1])
                del offsets[i]
                range_i -= 1
            else:
                i = i + 1

    def _sum_vect(self, dic1, dic2):
        """
        DESCRIPTION: Adding two vectors in form of dictionaries (sparse vectors or inverted list)
        INPUTS: dic1 <dictionary> - Vector 1
                dic2 <dictionary> - Vector 2
        OUTPUT: res <dictionary> - Sum of the two vectors
        """
        res = copy.deepcopy(dic1)
        for i in dic2.keys():
            if i in res:
                res[i] += dic2[i]
            else:
                res[i] = dic2[i]
        return res

    def _tf_isf(self, list_dic1, voc1, list_dic2, voc2):
        """
        DESCRIPTION: Compute the tf-idf <tf x ln(N/df)>  from a list of sentences with tf and the vocabularies in
        suspicios and source document
        INPUT: list_dic1 <list of dictionaries> -  List containing the vectors (dictionaries) of document 1
               voc1 <dictionary> - Vocabulary at document 1 with the idf of each one
               list_dic2 <list of dictionaries> -  List containing the vectors (dictionaries) of document 2
               voc2 <dictionary> - Vocabulary at document 2 with the idf of each one
        OUTPUT: No returned value. Modify the inputs list_dic1 and list_dic2
        """
        df = self._sum_vect(voc1, voc2)
        td = len(list_dic1) + len(list_dic2)
        for i in range(len(list_dic1)):
            for j in list_dic1[i].keys():
                list_dic1[i][j] *= math.log(td / float(df[j]))
        for i in range(len(list_dic2)):
            for j in list_dic2[i].keys():
                list_dic2[i][j] *= math.log(td / float(df[j]))



from __future__ import annotations

from dataclasses import dataclass

from lxml import etree
from lxml.builder import E

from plagdef.model.legacy.algorithm import DocumentPairMatches
from plagdef.model.preprocessing import Document


@dataclass(frozen=True)
class DocumentPairReport:
    doc1: Document
    doc2: Document
    content: str
    format: str


def generate_text_report(matches: list[DocumentPairMatches]) -> str:
    doc_pair_reports = []
    for doc_pair_matches in matches:
        doc1, doc2 = doc_pair_matches.doc1, doc_pair_matches.doc2
        report = f'Pair({doc1.name}, {doc2.name}):\n'
        for match in doc_pair_matches.list():
            report += f'  Match(Section({match.sec1.offset}, {match.sec1.length}), Section({match.sec2.offset},' \
                      f' {match.sec2.length}))\n'

        doc_pair_reports.append(DocumentPairReport(doc1, doc2, report, 'txt'))
    intro = 'There are no matching text sections in given documents.'
    if doc_pair_reports:
        intro = 'Reporting matches for each pair like this:\n' \
                f'  Match(Section(offset, length), Section(offset, length))\n\n'
    return intro + ''.join(f'{doc_pair_report.content}' for doc_pair_report in doc_pair_reports)


def generate_xml_reports(matches: list[DocumentPairMatches]) -> list[DocumentPairReport]:
    doc_pair_reports = []
    for doc_pair_matches in matches:
        root = E.report(doc1=doc_pair_matches.doc1.name, doc2=doc_pair_matches.doc2.name)
        for match in doc_pair_matches.list():
            root.append(E.match(
                doc1_offset=str(match.sec1.offset), doc1_length=str(match.sec1.length),
                doc2_offset=str(match.sec2.offset), doc2_length=str(match.sec2.length)
            ))
        report = etree.tostring(root, xml_declaration=True, encoding='UTF-8', pretty_print=True).decode('utf8')
        doc_pair_reports.append(DocumentPairReport(doc_pair_matches.doc1, doc_pair_matches.doc2, report, 'xml'))
    return doc_pair_reports



from __future__ import annotations

import pathlib
from functools import partial

import pkg_resources
from click import UsageError

from plagdef import services
from plagdef.model.legacy.algorithm import DocumentPairMatches
from plagdef.repositories import ConfigFileRepository, DocumentFileRepository, DocumentPairReportFileRepository, \
    UnsupportedFileFormatError, NoDocumentFilePairFoundError

CONFIG_PATH = pathlib.Path('config/alg.ini')


def find_matches(doc_repo=None, config_repo=None):
    if not config_repo:
        config_path = pkg_resources.resource_filename(__name__, str(CONFIG_PATH))
        config_repo = ConfigFileRepository(pathlib.Path(config_path))
    if not doc_repo:
        return partial(_needs_doc_dir_and_lang, config_repo=config_repo)
    else:
        return partial(services.find_matches, (doc_repo, config_repo))


def _needs_doc_dir_and_lang(doc_dir: str, lang: str, config_repo, recursive=False):
    try:
        doc_repo = DocumentFileRepository(pathlib.Path(doc_dir), lang, recursive)
        return services.find_matches(doc_repo, config_repo)
    except (NotADirectoryError, UnsupportedFileFormatError, NoDocumentFilePairFoundError) as e:
        raise UsageError(str(e)) from e


def write_xml_reports(doc_pair_report_repo=None):
    if not doc_pair_report_repo:
        return _needs_matches_and_xml_dir
    else:
        return partial(services.write_xml_reports, doc_pair_report_repo=doc_pair_report_repo)


def _needs_matches_and_xml_dir(matches: list[DocumentPairMatches], xml_dir: str):
    try:
        doc_pair_report_repo = DocumentPairReportFileRepository(pathlib.Path(xml_dir))
        return services.write_xml_reports(matches, doc_pair_report_repo)
    except NotADirectoryError as e:
        raise UsageError(str(e)) from e



from __future__ import annotations

from configparser import ParsingError

from click import UsageError

from plagdef.model import reporting
from plagdef.model.legacy import algorithm
from plagdef.model.legacy.algorithm import InvalidConfigError
from plagdef.model.preprocessing import UnsupportedLanguageError


def find_matches(doc_repo, config_repo) -> list[algorithm.DocumentPairMatches]:
    try:
        docs = doc_repo.list()
        config = config_repo.get()
        doc_pair_matches = algorithm.find_matches(docs, doc_repo.lang, config)
        return doc_pair_matches
    except (ParsingError, InvalidConfigError, UnsupportedLanguageError) as e:
        raise UsageError(str(e)) from e


def write_xml_reports(matches: list[algorithm.DocumentPairMatches], doc_pair_report_repo):
    doc_pair_reports = reporting.generate_xml_reports(matches)
    for report in doc_pair_reports:
        doc_pair_report_repo.add(report)



from __future__ import annotations

import math
from collections import Counter
from dataclasses import dataclass

import stanza
from more_itertools import pairwise

from plagdef.model import stopwords


class Preprocessor:
    def __init__(self, lang: str, min_sent_len: int, rem_stop_words: bool):
        self._lang = lang
        self._min_sent_len = min_sent_len
        self._rem_stop_words = rem_stop_words
        if lang == 'eng':
            try:
                self._nlp_model = stanza.Pipeline('en', processors='tokenize,mwt,pos,lemma', logging_level='WARN')
            except:  # Unpickling error raises Exception, cannot narrow
                stanza.download('en', processors='tokenize,mwt,pos,lemma', logging_level='INFO')
                self._nlp_model = stanza.Pipeline('en', processors='tokenize,mwt,pos,lemma', logging_level='WARN')
        elif lang == 'ger':
            try:
                self._nlp_model = stanza.Pipeline('de', processors='tokenize,mwt,pos,lemma', logging_level='WARN')
            except:  # Unpickling error raises Exception, cannot narrow
                stanza.download('de', processors='tokenize,mwt,pos,lemma', logging_level='INFO')
                self._nlp_model = stanza.Pipeline('de', processors='tokenize,mwt,pos,lemma', logging_level='WARN')
        else:
            raise UnsupportedLanguageError(f'The language "{lang}" is currently not supported.')

    # TODO: Only for compatibility
    def set_docs(self, doc1: Document, doc2: Document):
        self._doc1 = doc1
        self._doc2 = doc2

    # TODO: Only for compatibility
    def preprocess(self, legacy_obj):
        self.preprocess_doc_pair(self._doc1, self._doc2)
        legacy_obj.susp_voc = self._doc1.vocab
        legacy_obj.susp_bow = [sent.bow_tf_isf for sent in self._doc1.sents]
        legacy_obj.susp_offsets = [(sent.start_char, sent.end_char - sent.start_char) for sent in self._doc1.sents]
        legacy_obj.src_voc = self._doc2.vocab
        legacy_obj.src_bow = [sent.bow_tf_isf for sent in self._doc2.sents]
        legacy_obj.src_offsets = [(sent.start_char, sent.end_char - sent.start_char) for sent in self._doc2.sents]

    # TODO: New single preprocess
    def preprocess_new(self, docs: list[Document]):
        parsed_docs = self._nlp_model([stanza.Document([], text=doc.text) for doc in docs]) if docs else []
        stop_words = stopwords.ENGLISH if self._lang == 'eng' else stopwords.GERMAN
        for doc_idx, parsed_doc in enumerate(parsed_docs):
            for sent in parsed_doc.sentences:
                if self._rem_stop_words:
                    sent_lemmas = [word.lemma for word in sent.words
                                   if not word.upos == 'PUNCT' and word.text.lower() not in stop_words]
                else:
                    sent_lemmas = [word.lemma for word in sent.words if not word.upos == 'PUNCT']
                if len(sent_lemmas):
                    lemma_count = Counter(sent_lemmas)
                    docs[doc_idx].sents.append(
                        Sentence(docs[doc_idx], sent.tokens[0].start_char, sent.tokens[-1].end_char, lemma_count, {}))
                    for lemma in lemma_count.keys():
                        docs[doc_idx].vocab[lemma] += 1

            self._join_small_sentences(docs[doc_idx])

    # TODO: New pair preprocess function
    def preprocess_doc_pair(self, doc1: Document, doc2: Document):
        """
        Compute the tf-isf = tf x ln(N/sf), N being the number of sentences in corpus
        and sf the number of sentences containing the term
        """
        sf = doc1.vocab + doc2.vocab
        N = len(doc1.sents) + len(doc2.sents)
        for sent in doc1.sents:
            for lemma in sent.bow:
                sent.bow_tf_isf[lemma] = sent.bow[lemma] * math.log(N / float(sf[lemma]))
        for sent in doc2.sents:
            for lemma in sent.bow:
                sent.bow_tf_isf[lemma] = sent.bow[lemma] * math.log(N / float(sf[lemma]))

    def _join_small_sentences(self, doc: Document):
        for sent1, sent2 in pairwise(doc.sents):
            if sum(sent1.bow.values()) < self._min_sent_len or (
                sent2 == doc.sents[-1] and sum(sent2.bow.values()) < self._min_sent_len):
                for lemma in sent1.bow.keys():
                    if lemma in sent2.bow:
                        doc.vocab[lemma] -= 1
                sent2.bow.update(sent1.bow)
                sent2.start_char = sent1.start_char
                doc.sents.remove(sent1)


class Document:
    def __init__(self, name: str, text: str):
        self.name = name
        self.text = text
        self.vocab = Counter()  # <lemma, sent_freq>
        self.sents: list[Sentence] = []

    def __eq__(self, other):
        return isinstance(other, Document) and self.name == other.name

    def __hash__(self):
        return hash(self.name)


@dataclass
class Sentence:
    doc: Document
    start_char: int
    end_char: int
    bow: Counter
    bow_tf_isf: dict


class UnsupportedLanguageError(Exception):
    pass
